/* Nabi - X Input Method server for hangul
 * Copyright (C) 2003 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <wchar.h>
#include "nabi_keysym.h"
#include "server.h"
#include "nabi.h"

static
uint32_t string_to_hex(char* p)
{
    uint32_t ret = 0;
    uint32_t remain = 0;

    if (*p == 'U')
	p++;

    while (*p != '\0') {
	if (*p >= '0' && *p <= '9')
	    remain = *p - '0';
	else if (*p >= 'a' && *p <= 'f')
	    remain = *p - 'a' + 10;
	else if (*p >= 'A' && *p <= 'F')
	    remain = *p - 'A' + 10;
	else
	    return 0;

	ret = ret * 16 + remain;
	p++;
    }
    return ret;
}

#define COMPOSE_BUF_MAX 256
static int
load_compose_map_from_file(char * filename, NabiComposeMap * compose_map)
{
	FILE * file;
	int i;
	char *line, *p, *saved_position;
	char buf[COMPOSE_BUF_MAX];
	uint32_t key1, key2;
	wchar_t value;
	NabiComposeItem *citem;

	/* init */
	compose_map->name = NULL;
	for(i = 0; i < COMPOSE_MAP_SIZE; i++)
		compose_map->map[i] = NULL;
	compose_map->size = 0;
	file = fopen(filename, "r");
	if(file == NULL) {
		fprintf(stderr, "Nabi: Can't read compose map file\n");
		return 0;
	}

	i = 0;
	for (line = fgets(buf, sizeof(buf), file);
		line != NULL;
		line = fgets(buf, sizeof(buf), file)) {
		p = strtok_r(line, " \t\n", &saved_position);
		/* comment */
		if (p == NULL || p[0] == '#')
	   		continue;

		if (strcmp(p, "Name:") == 0) {
			p = strtok_r(NULL, "\n", &saved_position);
			if (p == NULL)
				continue;
			compose_map->name = strdup(p);
			continue;
		} else {
			key1 = string_to_hex(p);
			if (key1 == 0)
				continue;

			p = strtok_r(NULL, " \t", &saved_position);
			if (p == NULL)
				continue;
			key2 = string_to_hex(p);
			if (key2 == 0)
				continue;

			p = strtok_r(NULL, " \t", &saved_position);
			if (p == NULL)
				continue;
			value = string_to_hex(p);
			if (value == 0)
				continue;

			citem = (NabiComposeItem*)
				malloc(sizeof(NabiComposeItem));
			citem->key = key1 << 16 | key2;
			citem->code = value;

			compose_map->map[i] = citem;
			i++;
		}
	}
	fclose(file);

	if(compose_map->name == NULL) {
		for(i--; i < 0; i--)
			free(compose_map->map[i]);
		return 0;
	}

	/* FIX ME: no list sorting */
	compose_map->size = i;
	return 1;
}

static NabiKeyboardMap *
load_keyboard_map_from_file(const char * filename)
{
	FILE * file;
	int i;
	char *line, *p, *saved_position;
	char buf[256];
	wchar_t key, value;
	NabiKeyboardMap * keyboard_map;

	file = fopen(filename, "r");
	if(file == NULL) {
		fprintf(stderr, "Nabi: Can't read keyboard map file\n");
		return NULL;
	}

	keyboard_map = malloc(sizeof(NabiKeyboardMap));

	/* init */
	keyboard_map->type = NABI_KEYBOARD_3SET;
	keyboard_map->filename = strdup(filename);
	keyboard_map->name = NULL;
	for(i = 0; i < sizeof(keyboard_map->map) / sizeof(keyboard_map->map[0]); i++)
		keyboard_map->map[i] = 0;

	for(line = fgets(buf, sizeof(buf), file);
	    line != NULL;
	    line = fgets(buf, sizeof(buf), file)) {
		p = strtok_r(line, "\t\n", &saved_position);
		/* comment */
		if(p == NULL || p[0] == '#')
			continue;
		if(strcmp(p, "Name:") == 0) {
			p = strtok_r(NULL, "\n", &saved_position);
			if(p == NULL)
				continue;
			keyboard_map->name = strdup(p);
			continue;
		}
		else if(strcmp(p, "Type2") == 0) {
			keyboard_map->type = NABI_KEYBOARD_2SET;
		}
		else {
			key = string_to_hex(p);
			if(key == 0)
				continue;
			p = strtok_r(NULL, "\t", &saved_position);
			if(p == NULL)
				continue;
			value = string_to_hex(p);
			if(value == 0)
				continue;

			if(key < XK_exclam || key > XK_asciitilde)
				continue;

			keyboard_map->map[key - XK_exclam] = value;
		}
	}
	fclose(file);

	if(keyboard_map->name == NULL)
		keyboard_map->name = basename(keyboard_map->filename);

	return keyboard_map;
}

void
load_compose_map(char * filename)
{
	NabiComposeMap * compose_map;

	compose_map = (NabiComposeMap *)malloc(sizeof(NabiComposeMap));
	load_compose_map_from_file(filename, compose_map);
	nabi_server_set_compose_map(nabi_server, compose_map->map,
						 compose_map->size);
}

void
set_up_keyboard(const char * filename, NabiKeyboardType type)
{
	const wchar_t * keyboard_map;

	keyboard_map = load_keyboard_map_from_file(filename)->map;
	if(keyboard_map == NULL)
		fprintf(stderr, "Nabi: Error - No keyboard map\n");
	else
		nabi_server_set_keyboard(nabi_server,
					 keyboard_map,
					 type);
}

void
nabi_app_init(char * type)
{
	if(!strcmp(type, "3") || !strcmp(type, "3final")) {
		load_compose_map(COMPOSE_MAP_DEFAULT);
		set_up_keyboard(KBD_3FQWERTY, NABI_KEYBOARD_3SET);
	}
	else if(!strcmp(type, "390")) {
		load_compose_map(COMPOSE_MAP_DEFAULT);
		set_up_keyboard(KBD_39QWERTY, NABI_KEYBOARD_3SET);
	}
	else {
		load_compose_map(COMPOSE_MAP_DEFAULT);
		set_up_keyboard(KBD_2QWERTY, NABI_KEYBOARD_2SET);
	}
}
